// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package main

import (
	"bytes"
	"context"
	"encoding/base64"
	"encoding/hex"
	"fmt"
	"io"
	"net/http"
	"os"
	"strings"
	"testing"
	"time"

	"hello-api/internal/app/middleware"

	"carbonblack.com/protocols/psc/devicelastcontacttime"

	"hello-api/internal/config"
	"hello-api/internal/logger"
	"hello-api/internal/utils"

	"github.com/aws/aws-sdk-go-v2/aws"
	"github.com/aws/aws-sdk-go-v2/service/kinesis"
	"github.com/aws/aws-sdk-go-v2/service/kinesis/types"

	"hello-api/internal/app/controllers"
	"hello-api/internal/app/parsers/deviceresponseheader"

	deviceauthpb "carbonblack.com/protocols/psc/ems/server/deviceauth"
	. "github.com/onsi/ginkgo/v2"
	. "github.com/onsi/gomega"
	deagg "github.com/zacharyestep/kinesis-producer/deaggregation"
	"google.golang.org/protobuf/proto"
)

const helloAPIRoute = "services/hello"

func getEndpoint() string {
	env, found := os.LookupEnv("ENV")      // flag; define to mark as local run
	host := "http://ems-device-hello:5004" // port 80 under normal running, 5004 when local testing
	if found && strings.ToLower(env) == "local" {
		host = "http://localhost:5004"
	}
	return fmt.Sprintf("%s/%s", host, helloAPIRoute)
}

//goland:noinspection GoUnhandledErrorResult
var _ = Describe("functional tests", func() {
	const (
		orgKey             = "7ZP4H5XR" // ENV local, OrgId 2023
		orgId              = 2023
		deviceId           = 123
		deviceGroupId      = 456
		defaultStreamName  = "local-ems-device-hello-local-infra-lct-stream-psc-ems-last-contact-time-updater-stream"
		defaultAwsRegion   = "us-east-1"
		defaultAwsEndpoint = "http://localhost:4566"

		// Encoded:
		//  - version: 17
		//  - lastUpdateTime: Sunday, February 2, 2020 12:00:00 GMT, 1580644800000
		//  - lastLiveQueryTime: Monday, September 9, 2019 12:00:00 GMT, 1568030400000
		encodedHeaderString = "c011c0c0c0c0c0c0c0c0c0d30000017005c5960081b14c6173744c697665517565727954696d65cf0000016d15e51e00"
	)

	var (
		httpClient = http.DefaultClient

		lctStreamName = utils.EnvOrDefault("LCT_STREAM_NAME", defaultStreamName)
		aswEndpoint   = utils.EnvOrDefault("AWS_ENDPOINT", defaultAwsEndpoint)
		region        = utils.EnvOrDefault("REGION", defaultAwsRegion)
	)

	testContextHeader := &deviceauthpb.Context{
		Device: &deviceauthpb.Device{
			Id: deviceId, DeviceType: deviceauthpb.DeviceType_LINUX,
			DeploymentType: deviceauthpb.DeploymentType_ENDPOINT, OrgKey: orgKey,
		},
		DeviceGroup:  &deviceauthpb.DeviceGroup{Id: deviceGroupId, OrgId: orgId},
		Organization: &deviceauthpb.Organization{OrgKey: orgKey, Id: orgId},
	}
	logger.BaseLogger().Debug(fmt.Sprintf("Device Context: %+v", testContextHeader))

	marshalledHeader, _ := proto.Marshal(testContextHeader)
	encodedHeader := base64.StdEncoding.EncodeToString(marshalledHeader)

	// NOTE: negative tests beforehand to ensure they don't bring down the service!

	It("Validates hello API operation with missing device context", func() {
		logger.BaseLogger().Info(utils.HeaderStr("Validate Missing Context Handling"))
		decodedBuffer, _ := hex.DecodeString(encodedHeaderString)
		request, err := http.NewRequest("POST", getEndpoint(), bytes.NewBuffer(decodedBuffer))
		if err != nil {
			logger.BaseLogger().Error(fmt.Sprintf("Failed to get new request: %v", err))
		}
		Expect(err).To(BeNil())
		Expect(request).ToNot(BeNil())

		logger.BaseLogger().Info(fmt.Sprintf("Calling httpClient Do() to %s", getEndpoint()))
		logger.BaseLogger().Info(fmt.Sprintf("Msgpack: %s", decodedBuffer))
		response, err := httpClient.Do(request)
		if err != nil {
			logger.BaseLogger().Error(fmt.Sprintf("Failed to make POST call: %v", err))
		}
		Expect(err).To(BeNil())
		defer response.Body.Close()

		// Check 1: Hello Response Code
		Expect(response).ToNot(BeNil())
		logger.BaseLogger().Info(fmt.Sprintf("RESPONSE: %+v", response))
		Expect(response.StatusCode).To(BeComparableTo(http.StatusInternalServerError))
	})

	It("Validates hello API operation with bogus device context", func() {
		logger.BaseLogger().Info(utils.HeaderStr("Validate Bogus Context Handling"))
		decodedBuffer, _ := hex.DecodeString(encodedHeaderString)
		request, err := http.NewRequest("POST", getEndpoint(), bytes.NewBuffer(decodedBuffer))
		if err != nil {
			logger.BaseLogger().Error(fmt.Sprintf("Failed to get new request: %v", err))
		}
		Expect(err).To(BeNil())
		Expect(request).ToNot(BeNil())

		logger.BaseLogger().Info(fmt.Sprintf("Calling httpClient Do() to %s", getEndpoint()))
		logger.BaseLogger().Info(fmt.Sprintf("Msgpack: %s", decodedBuffer))
		logger.BaseLogger().Info(fmt.Sprintf("Device Context: %s", "BOGUS"))
		request.Header.Set(middleware.DeviceContextHeader, "BOGUS")
		response, err := httpClient.Do(request)
		if err != nil {
			logger.BaseLogger().Error(fmt.Sprintf("Failed to make POST call: %v", err))
		}
		Expect(err).To(BeNil())
		defer response.Body.Close()

		// Check 1: Hello Response Code
		Expect(response).ToNot(BeNil())
		logger.BaseLogger().Info(fmt.Sprintf("RESPONSE: %+v", response))
		Expect(response.StatusCode).To(BeComparableTo(http.StatusInternalServerError))
	})

	It("Validates hello API operation with missing device msgpack", func() {
		logger.BaseLogger().Info(utils.HeaderStr("Validate missing device msgpack"))
		request, err := http.NewRequest("POST", getEndpoint(), http.NoBody)
		if err != nil {
			logger.BaseLogger().Error(fmt.Sprintf("Failed to get new request: %v", err))
		}
		Expect(err).To(BeNil())
		Expect(request).ToNot(BeNil())

		request.Header.Set(middleware.DeviceContextHeader, encodedHeader)
		logger.BaseLogger().Info(fmt.Sprintf("Calling httpClient Do() to %s", getEndpoint()))
		logger.BaseLogger().Info(fmt.Sprintf("Device Context: %v (%s)", testContextHeader, encodedHeader))
		response, err := httpClient.Do(request)
		if err != nil {
			logger.BaseLogger().Error(fmt.Sprintf("Failed to make POST call: %v", err))
		}
		Expect(err).To(BeNil())
		defer response.Body.Close()

		// Check 1: Hello Response Code
		Expect(response).ToNot(BeNil())
		logger.BaseLogger().Info(fmt.Sprintf("RESPONSE: %+v", response))
		Expect(response.StatusCode).To(BeComparableTo(http.StatusInternalServerError))
	})

	It("Validates hello API operation with bogus device msgpack", func() {
		logger.BaseLogger().Info(utils.HeaderStr("Validate bogus device msgpack"))
		decodedBuffer, _ := hex.DecodeString("DEADBEEF")
		request, err := http.NewRequest("POST", getEndpoint(), bytes.NewBuffer(decodedBuffer))
		if err != nil {
			logger.BaseLogger().Error(fmt.Sprintf("Failed to get new request: %v", err))
		}
		Expect(err).To(BeNil())
		Expect(request).ToNot(BeNil())

		request.Header.Set(middleware.DeviceContextHeader, encodedHeader)
		logger.BaseLogger().Info(fmt.Sprintf("Calling httpClient Do() to %s", getEndpoint()))
		logger.BaseLogger().Info(fmt.Sprintf("Msgpack: %s", decodedBuffer))
		logger.BaseLogger().Info(fmt.Sprintf("Device Context: %s", "BOGUS"))
		response, err := httpClient.Do(request)
		if err != nil {
			logger.BaseLogger().Error(fmt.Sprintf("Failed to make POST call: %v", err))
		}
		Expect(err).To(BeNil())
		defer response.Body.Close()

		// Check 1: Hello Response Code
		Expect(response).ToNot(BeNil())
		logger.BaseLogger().Info(fmt.Sprintf("RESPONSE: %+v", response))
		Expect(response.StatusCode).To(BeComparableTo(http.StatusInternalServerError))
	})

	It("Validates hello API operation and response", func() {
		logger.BaseLogger().Info(utils.HeaderStr("Validate hello API response"))
		decodedBuffer, _ := hex.DecodeString(encodedHeaderString)
		request, err := http.NewRequest("POST", getEndpoint(), bytes.NewBuffer(decodedBuffer))
		if err != nil {
			logger.BaseLogger().Error(fmt.Sprintf("Failed to get new request: %v", err))
		}
		Expect(err).To(BeNil())
		Expect(request).ToNot(BeNil())

		request.Header.Set(middleware.DeviceContextHeader, encodedHeader)
		logger.BaseLogger().Info(fmt.Sprintf("Calling httpClient Do() to %s", getEndpoint()))
		logger.BaseLogger().Info(fmt.Sprintf("Msgpack: %s", decodedBuffer))
		logger.BaseLogger().Info(fmt.Sprintf("Device Context: %v (%s)", testContextHeader, encodedHeader))
		response, err := httpClient.Do(request)
		if err != nil {
			logger.BaseLogger().Error(fmt.Sprintf("Failed to make POST call: %v", err))
		}
		Expect(err).To(BeNil())
		defer response.Body.Close()

		// Check 1: Hello Response Code
		Expect(response).ToNot(BeNil())
		Expect(response.StatusCode).To(BeComparableTo(http.StatusOK))

		decoder := deviceresponseheader.NewResponseHeaderDecoder()
		var temp strings.Builder
		_, err = io.Copy(&temp, response.Body)
		if err != nil {
			logger.BaseLogger().Error(fmt.Sprintf("Unable to copy response body from message: %v", err))
		}
		Expect(err).To(BeNil())
		responseBodyParsed, err := decoder.Decode(temp.String())
		if err != nil {
			logger.BaseLogger().Error(fmt.Sprintf("Unable to decode response body: %v", err))
		}
		Expect(err).To(BeNil())

		// Check 2: Validate Hello Response
		logger.BaseLogger().Info(fmt.Sprintf("RESPONSE: %+v", responseBodyParsed))

		// NOTE: Local testing uses a mock service that always returns the same values
		// Version in the response is same as the version received in the request
		Expect(responseBodyParsed.Version).To(BeEquivalentTo(17))
		Expect(responseBodyParsed.Code).To(BeEquivalentTo(controllers.HelloResponseCode))
		Expect(responseBodyParsed.Success).To(BeEquivalentTo(true))
		Expect(responseBodyParsed.ErrorCode).To(BeEquivalentTo(0))
		Expect(responseBodyParsed.Message).To(BeEquivalentTo("Success."))
		Expect(responseBodyParsed.Hints).To(BeEquivalentTo(MockHints))
		Expect(responseBodyParsed.LastUpdateTime).To(BeEquivalentTo(&MockUpdateTime))
		Expect(responseBodyParsed.LastLiveQueryTime).To(BeEquivalentTo(&MockQueryTime))
	})

	It("Validates hello push to LCT stream", func() {
		logger.BaseLogger().Info(utils.HeaderStr("Validate hello push to LCT stream"))
		kinesisClient, err := config.CreateKinesisClient(&config.Config{Region: region, EnvCategory: "local", AwsEndpoint: aswEndpoint})
		if err != nil {
			logger.BaseLogger().Error(fmt.Sprintf("Failed to create client: %v", err))
		}
		Expect(err).To(BeNil())

		shards, err := kinesisClient.ListShards(context.TODO(), &kinesis.ListShardsInput{StreamName: aws.String(lctStreamName)})
		if err != nil {
			logger.BaseLogger().Error(fmt.Sprintf("Failed to get list of shards: %v", err))
		}
		Expect(err).To(BeNil())

		shardName := shards.Shards[0]
		shardIterator, err := kinesisClient.GetShardIterator(context.TODO(), &kinesis.GetShardIteratorInput{
			ShardId: shardName.ShardId, ShardIteratorType: types.ShardIteratorTypeLatest,
			StreamName: aws.String(lctStreamName),
		})
		if err != nil {
			logger.BaseLogger().Error(fmt.Sprintf("Failed to get shard iterator: %v", err))
		}
		Expect(err).To(BeNil())

		// Check 1: Ensure that LCT stream had records added
		var recordOutput *kinesis.GetRecordsOutput = nil
		timeout := 60
		for (recordOutput == nil || len(recordOutput.Records) == 0) && timeout > 0 {
			input := kinesis.GetRecordsInput{ShardIterator: shardIterator.ShardIterator}
			recordOutput, err = kinesisClient.GetRecords(context.TODO(), &input)
			if err != nil {
				logger.BaseLogger().Error(fmt.Sprintf("Failed to get records: %v", err))
			}
			Expect(err).To(BeNil())
			if len(recordOutput.Records) == 0 {
				time.Sleep(time.Second) // allow for LCT propagation
				timeout--
			}
		}
		Expect(len(recordOutput.Records)).ToNot(BeEquivalentTo(0))

		// Check 2: Ensure that added records correspond to created device
		records, err := deagg.Unmarshal(recordOutput.Records[0].Data)
		if err != nil {
			logger.BaseLogger().Error(fmt.Sprintf("Failed to deaggregate records: %v", err))
		}
		Expect(err).To(BeNil())

		recordFound := false
		for _, record := range records.Records {
			lct := devicelastcontacttime.DeviceLastContactTime{}
			err := proto.Unmarshal(record.Data, &lct)
			if err != nil {
				logger.BaseLogger().Error(fmt.Sprintf("Failed to unmarshal data: %v", err))
			}
			Expect(err).To(BeNil())

			recordFound = true

			//goland:noinspection GoVetCopyLock
			logger.BaseLogger().Info(fmt.Sprintf("LCT Data Seen: OrgKey=%s, DevId=%d, LctTime=%d, DeRegTime=%d",
				lct.OrgKey, lct.DeviceId, lct.LastContactTimestamp, lct.DeregisterAfterTime))

			Expect(lct.DeviceId).To(BeEquivalentTo(deviceId))
			Expect(lct.OrgKey).To(BeEquivalentTo(orgKey))
			Expect(lct.LastContactTimestamp).ToNot(BeEquivalentTo(0)) // will be "now"
			logger.BaseLogger().Info(fmt.Sprintf("LctTime (now) resolves to: %s",
				time.UnixMilli(lct.LastContactTimestamp).Format(time.RFC3339)))
			Expect(lct.DeregisterAfterTime).To(BeEquivalentTo(0))

		}
		Expect(recordFound).To(BeTrue())
	})
})

func TestFunctionals(t *testing.T) {
	RegisterFailHandler(Fail)
	RunSpecs(t, "Functional tests")
}
